%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 370-CMAC-B.SL - second half of patterns for IBM370 PSL cmacro expansion  %
%                                                                          %
% Author:      Jim McDonald                                                %
%              IMSSS                                                       %
%              Stanford University                                         %
% Date:        11 December 1982                                            %
%                                                                          %
% Based on HP-CMAC.SL of Mar 82 by Maguire, Griss, Lowder                  %
% Based on VAX-CMAC.SL of 14 Jan 1982 by Eric Benson                       %
%                                                                          %
% Revision:    27 June 1982 -- split into two files (-A and -B).           %
%                                                                          %
% All functions have been moved out of this file.                          %
%       Only patterns remain.                                              %
%       See 370-CMAC-FNS.SL for functions used here.                       %
%       See 370-CMAC-A.SL   for first set of patterns.                     %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                       COMPARISONS                                         %
%                                                                           %
% Compare  -- compares arg1 with arg2                                       %
%                                                                           %
% (See the note above about jumps.)                                         %
%                                                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%               AVAILABLE 370 INSTRUCTIONS                                  %
%                                                                           %
% These instuctions are all (and only) those that do "pure" comparisons,    %
%   i.e., they do not modify memory.                                        %
%   Some modify a register, and are marked as follows:                      %
%                                                                           %
%   *,  R1 is modified.                                                     %
%   *,* R1 and R2 are modified                                              %
%                                                                           %
% Normal comparisons:                                                       %
%                                                                           %
% =     CR   R1,R2           RR Compare                                     %
% =     C    R1,D2(X2,B2)    RX Compare         (32 bit signed integers)    %
% =     CH   R1,D2(X2,B2)    RX Compare Halfword                            %
%       CLR  R1,R2           RR Compare Logical (32 bit unsigned integers)  %
%       CL   R1,D2(X2,B2)    RX Compare Logical                             %
% =     CLI  D1(B1),I2       SI Compare Logical (immediate)                 %
% =     CLC  D1(L,B1),D2(B2) SS Compare Logical (up to 256 bytes)           %
% = *,* CLCL R1,R2           RR Compare Logical Long                        %
%       CLM  R1,M3,D2(B2)    RS Compare Logical Characters Under Mask       %
%       TM   D1(B1),I2       SI Test Under Mask                             %
%                                                                           %
% = *,  LTR  R1,R2           RR Load And Test                               %
% = *,  LCR  R1,R2           RR Load Complement                             %
%                                                                           %
% Floating point comparisons (R1,R2 refer to floating point regs):          %
%                                                                           %
%       CER  R1,R2           RR Compare (short normalized)                  %
%       CE   R1,D2(X2,B2)    RX Compare (short normalized)                  %
%       CDR  R1,R2           RR Compare (long normalized)                   %
%       CD   R1,D2(X2,B2)    RX Compare (long normalized)                   %
%   *,  LTER R1,R2           RR Load and test (short normalized)            %
%   *,  LTDR R1,R2           RR Load and test (long normalized)             %
%   *,  LCER R1,R2           RR Load Complement (short normalized)          %
%   *,  LCDR R1,R2           RR Load Complement (long normalized)           %
%                                                                           %
% Decimal comparison:                                                       %
%                                                                           %
%       CP   D1(L1,B1),D2(L2,B2)   SS Compare decimal                       %
%                                                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Process Interlocks (these modify memory):                                 %
%                                                                           %
%       CS  R1,R3,D2(B2)     RS Compare and Swap                            %
%       CDS R1,R3,D2(B2)     RS Compare Double and Swap                     %
%                                                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
(DefCMacro !*Compare
        % assume that BigOffsetP implies FullWordP

        % special cases
        ((RegP       ZeroP)     (LTR ArgOne ArgOne)
                                (SameLineComment "Is" ArgOne "zero?"))
        % register -- register
        ((RegP       RegP)      (CR  ArgOne ArgTwo)
                                (SameLineComment "Is" ArgOne "=" ArgTwo "?"))

        ((AnyP       RegP)      (PROBLEM "Compare" ArgOne ArgTwo))

        % register -- immediate
        ((RegP       HalfInumP) (CH  ArgOne (Lit (Halfword ArgTwo)))
                                (SameLineComment "Is" ArgOne "=" ArgTwo "?"))
        %  CH does signed comparison, so can't use unsigned halfwords

        ((RegP       FullInumP) (C   ArgOne (Lit (Fullword ArgTwo)))
                                (SameLineComment "Is" ArgOne "=" ArgTwo "?"))
        ((RegP UnsignedFullInumP)
                                (C  ArgOne (Lit (Fullword (Hex ArgTwo))))
                                (SameLineComment "Is" ArgOne "=" ArgTwo "?"))

        % register -- memory
        ((RegP       HalfwordP) (CH  ArgOne ArgTwo)
                                (SameLineComment "Is" ArgOne "=" ArgTwo "?"))
        ((RegP      BigOffsetP) (!*Move (OffsetOf ArgTwo)
                                        (RegNotUsedBy ArgOne ArgTwo))
                                (C ArgOne
                                   (DoubleIndexed
                                        (RegNotUsedBy ArgOne ArgTwo)
                                        (BaseRegOf ArgTwo)
                                        0))
                                (SameLineComment "Is" ArgOne "=" ArgTwo "?"))
        ((RegP       FullwordP) (C   ArgOne ArgTwo)
                                (SameLineComment "Is" ArgOne "=" ArgTwo "?"))

        % memory -- immediate
        ((ByteP      ByteInumP) (CLI ArgOne ArgTwo)
                                (SameLineComment "Is" ArgOne "=" ArgTwo "?"))
%
% following could be a CLC if test is BE or BNE
% but JUMPIF doesn't pass that info -- hmmm.
% might also note that negative numbers are 40000000 to FFFFFFFF
% so we could hack up a tricky CLC when test is for fullword neg/nonneg
% by using 4000000 as comparator (sigh)
% in loops, tracking registers might reveal that desired value is already
% in a register, thus avoiding load (sigh)
%
        ((FullWordP  FullInumP) (!*Move ArgOne (RegNotUsedBy ArgTwo))
                                (!*Compare (RegNotUsedBy ArgTwo) ArgTwo))
                                                % CLC would ignore sign

        % immediate -- memory
        ((FullInumP  FullWordP) (!*Move ArgOne (RegNotUsedBy ArgTwo))
                                (!*Compare (RegNotUsedBy ArgTwo) ArgTwo))
                                                % CLC would ignore sign

        % memory -- memory
        ((ByteP      ByteP)     (!*SS1 CLC ArgOne ArgTwo 1))
        ((FullWordP  FullWordP) (!*Move ArgOne (RegNotUsedBy ArgTwo))
                                (!*Compare (RegNotUsedBy ArgTwo) ArgTwo))
                                                % CLC would ignore sign
        (NeqLengthP             (PROBLEM "Compare" ArgOne ArgTwo))
        ((Mem256P    Mem256P)   (!*SS1 CLC ArgOne ArgTwo nil))
        (LongMemP               (!*CLCL ArgOne ArgTwo))

        % We should have handled all cases by now
        ( (PROBLEM "Compare" ArgOne ArgTwo)))


(DefCMacro !*CLCL)   % Compare Characters Long -- needs 2 pairs of registers


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                       MOVES                                               %
%                                                                           %
%  Move          -- Move the object found at ArgOne to location ArgTwo.     %
%                                                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%               AVAILABLE 370 INSTRUCTIONS                                  %
%                                                                           %
%  = LR   R1,R2             RR Load                                         %
%  = L    R1,D2(X2,B2)      RX Load                                         %
%  = LH   R1,D2(X2,B2)      RX Load Halfword                                %
%  = IC   R1,D2(X2,B2)      RX Insert Character                             %
%  = LA   R1,D2(X2,B2)      RX Load Address                                 %
%    LM   R1,R3,D2(B2)      RS Load Multiple                                %
%    LPR  R1,D2(X2,B2)      RX Load Positive (abs value)     CC [FPOV]      %
%    LNR  R1,D2(X2,B2)      RX Load Negative (neg abs value) CC             %
%    ICM  R1,M3,D2(B2)      RS Insert Characters under Mask  CC             %
%  = LTR  R1,R2             RR Load and Test                 CC             %
%                                                                           %
%  = ST   R1,D2(X2,B2)      RX Store                                        %
%  = STH  R1,D2(X2,B2)      RX Store Halfword                               %
%  = STC  R1,D2(X2,B2)      RX Store Character                              %
%    STCM R1,M3,D2(B2)      RS Store Characters under Mask                  %
%    STM  R1,R3,D2(B2)      RS Store Multiple                               %
%                                                                           %
%  = MVI  D1(B1),I2           SI Move Immediate                             %
%  = MVC  D1(L,B1),D2(B2)     SS Move                                       %
%  = MVCL R1,R2               RR Move Long                   CC [SPEC]      %
%    MVN  D1(L,B1),D2(B2)     SS Move Numerics (low bits)                   %
%    MVZ  D1(L,B1),D2(B2)     SS Move Zones    (high bits)                  %
%    MVO  D1(L1,B1),D2(L2,B2) SS Move with Offset                           %
%                                                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
(DefCMacro !*Move            %  (!*Move Source Destination)
        % assume that BigOffsetP implies FullWordP

        % special cases
        (EqualP)
        ((ZeroP RegP)       (SR ArgTwo ArgTwo)
                            (SameLineComment ArgTwo ":= 0"))
        ((ZeroP BigOffsetP) (!*Move (OffsetOf ArgTwo) (RegNotUsedBy ArgTwo))
                            (SR (SecondRegNotUsedBy ArgTwo)
                                (SecondRegNotUsedBy ArgTwo))
                            (ST (SecondRegNotUsedBy ArgTwo)
                                (DoubleIndexed (BaseRegOf ArgTwo)
                                               (RegNotUsedBy ArgTwo)
                                               0))
                            (SameLineComment ArgTwo ":= 0"))
        ((ZeroP FullwordP)  (SR (RegNotUsedBy ArgTwo) (RegNotUsedBy ArgTwo))
                            (ST (RegNotUsedBy ArgTwo) ArgTwo)
                            (SameLineComment ArgTwo ":= 0"))
        ((ZeroP HalfwordP)  (SR (RegNotUsedBy ArgTwo) (RegNotUsedBy ArgTwo))
                            (STH (RegNotUsedBy ArgTwo) ArgTwo)
                            (SameLineComment ArgTwo ":= 0"))
        ((ZeroP     ByteP)  (SR (RegNotUsedBy ArgTwo) (RegNotUsedBy ArgTwo))
                            (STC (RegNotUsedBy ArgTwo) ArgTwo)
                            (SameLineComment ArgTwo ":= 0"))
                                % note: if some register was guaranteed to
                                %       have 0 in the low byte, we could use
                                %       it here instead of T1
        % register -> register
        ((RegP       RegP)  (LR ArgTwo ArgOne))

        % constant -> register
        ((DispInumP  RegP)  (LA ArgTwo ArgOne))

        ((HalfInumP  RegP)  (LH ArgTwo (Lit (Halfword ArgOne)))
                            (SameLineComment ArgTwo ":=" ArgOne))
        %  LH extends sign, so can't use unsigned halfwords

        ((HalfWordIdOffsetP  RegP)
                            (LH ArgTwo (IdOffsetLit (Halfword ArgOne)))
                            (SameLineComment ArgTwo ":=" ArgOne))

        ((FullInumP  RegP)  (L  ArgTwo (Lit (Fullword ArgOne)))
                            (SameLineComment ArgTwo ":=" ArgOne))
        ((UnsignedFullInumP  RegP)
                            (L  ArgTwo (Lit (Fullword (Hex ArgOne))))
                            (SameLineComment ArgTwo ":=" ArgOne))
        ((FullWordIdOffsetP  RegP)
                            (L  ArgTwo (IdOffsetLit (Fullword ArgOne)))
                            (SameLineComment ArgTwo ":=" ArgOne))

        % addresss -> register or addressible memory
        ((ImmediateP RegP)  (LA ArgTwo (unimmediate ArgOne)))
        ((ImmediateP BigOffsetP)
                            (!*Move (OffsetOf ArgTwo) (RegNotUsedBy ArgTwo))
                            (LA (SecondRegNotUsedBy ArgTwo)
                                (unimmediate Argone))
                            (ST (SecondRegNotUsedBy ArgTwo)
                                (DoubleIndexed (BaseRegOf ArgTwo)
                                               (RegNotUsedBy ArgTwo)
                                               0))
                            (SameLineComment ArgTwo ":=" ArgOne))
        ((ImmediateP FullwordP)
                            (LA (RegNotUsedBy ArgTwo) (unimmediate Argone))
                            (ST (RegNotUsedBy ArgTwo) ArgTwo)
                            (SameLineComment ArgTwo ":=" ArgOne))
        ((ImmediateP AnyP)  (PROBLEM "Move" ArgOne ArgTwo))


        % directly addressible memory -> register
        ((ByteP      RegP)   (SR ArgTwo ArgTwo)
                             (IC ArgTwo ArgOne))
        ((HalfWordP  RegP)   (LH ArgTwo ArgOne))
        ((BigOffsetP RegP)   (!*Move (OffsetOf ArgOne) (RegNotUsedBy ArgOne))
                             (L ArgTwo (DoubleIndexed (BaseRegOf ArgOne)
                                                      (RegNotUsedBy ArgOne)
                                                      0))
                             (SameLineComment ArgTwo ":=" ArgOne))
        ((FullWordP  RegP)   (L  ArgTwo ArgOne)
                             (SameLineComment ArgTwo ":=" ArgOne))
        ((AnyP       RegP)   (PROBLEM "Move" ArgOne ArgTwo))

        % register -> addressible memory
        ((RegP   ByteP)      (STC ArgOne ArgTwo))
        ((RegP   HalfwordP)  (STH ArgOne ArgTwo))
        ((RegP   BigOffsetP) (!*Move (OffsetOf ArgTwo)
                                     (RegNotUsedBy ArgOne ArgTwo))
                             (ST ArgOne
                                 (DoubleIndexed (BaseRegOf ArgTwo)
                                                (RegNotUsedBy ArgOne ArgTwo)
                                                0))
                             (SameLineComment ArgTwo ":=" ArgOne))
        ((RegP   FullwordP)  (ST  ArgOne ArgTwo)
                             (SameLineComment ArgTwo ":=" ArgOne))
        ((RegP   AnyP)       (PROBLEM "Move" ArgOne ArgTwo))

        % constant  -> addressible memory
        ((ByteInumP  ByteP)     (MVI ArgTwo ArgOne))
        ((HalfInumP  HalfWordP) (!*SS1 MVC ArgTwo (Lit (Halfword ArgOne)) 2))
        ((UnsignedHalfInumP  HalfWordP)
                                (!*SS1 MVC ArgTwo
                                       (Lit (Halfword (Hex ArgOne))) 2)
                                (SameLineComment ArgTwo ":=" ArgOne))
        ((FullInumP  BigOffsetP)
                            (!*Move (OffsetOf ArgTwo) (RegNotUsedBy ArgTwo))
                            (!*Move ArgOne (SecondRegNotUsedBy ArgTwo))
                            (ST (SecondRegNotUsedBy ArgTwo)
                                (DoubleIndexed (RegNotUsedBy ArgTwo)
                                               (BaseRegOf ArgTwo)
                                               0))
                            (SameLineComment ArgTwo ":=" ArgOne))
        ((FullInumP  FullWordP) (!*SS1 MVC ArgTwo (Lit (Fullword ArgOne)) 4)
                                (SameLineComment ArgTwo ":=" ArgOne))
        ((UnsignedFullInumP  FullWordP)
                                (!*SS1 MVC ArgTwo
                                       (Lit (Fullword (Hex ArgOne))) 4)
                                (SameLineComment ArgTwo ":=" ArgOne))

        % addressible memory -> addressible memory
        ((HalfWordP  HalfWordP)  (!*MaybeMVC ArgOne ArgTwo 2)
                                 (SameLineComment ArgTwo ":=" ArgOne))
        ((FullWordP  BigOffsetP) (!*Move ArgOne (RegNotUsedBy ArgTwo))
                                 (!*Move (RegNotUsedBy ArgTwo) ArgTwo))
        ((BigOffsetP FullWordP)  (!*Move ArgOne (RegNotUsedBy ArgTwo))
                                 (!*Move (RegNotUsedBy ArgTwo) ArgTwo))
        ((FullWordP  FullWordP)  (!*MaybeMVC ArgOne ArgTwo 4)
                                 (SameLineComment ArgTwo ":=" ArgOne))
        ((Mem256P    Mem256P)    (!*MaybeMVC ArgOne ArgTwo nil))
        (LongMemP                (!*MVCL ArgTwo ArgOne))

        % ??
        (                       (!*Move ArgOne (RegNotUsedBy ArgTwo))
                                (!*Move (RegNotUsedBy ArgTwo) ArgTwo)))

(DefCmacro !*MaybeMVC)
(DefCMacro !*MVCL)     % Move Characters Long -- needs 2 pairs of registers

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% MkItem may be used when evaluating WConst expressions.                %
% An item is formed on the 370 with the first eight bits as the Tag,    %
%   ----------------------------                                        %
%   | Tag  |   Info field      |                                        %
%   ----------------------------                                        %
%    0    7 8                31                                         %
%                                                                       %
% To create a constant ITEM, the TagPart can be inserted as the first   %
% byte of the object.  There should not be any need to AND and OR the   %
% arguments to achieve the desired result.                              %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(DefCMacro !*MkItem

        % building object in register
        ((RegP      ByteInumP)  (ICM ArgOne 2#1000 (Lit (Byte ArgTwo))))
        ((RegP      ByteP)      (ICM ArgOne 2#1000 ArgTwo))
        ((RegP      HalfWordP)  (ICM ArgOne 2#1000 (ByteLoc ArgTwo 1)))
        ((RegP      FullWordP)  (ICM ArgOne 2#1000 (ByteLoc ArgTwo 3)))
        ((RegP      RegP)       (!*WAnd ArgOne 16#FFFFFF)
                                (!*WShift ArgTwo 24)    % tag from low byte
                                (!*WOr ArgOne ArgTwo))

        % building object in memory
        ((FullWordP ByteInumP)  (MVI ArgOne ArgTwo))
        ((FullWordP ByteP)      (!*SS1 MVC ArgOne ArgTwo 1))
        ((FullWordP HalfWordP)  (!*SS1 MVC ArgOne (ByteLoc ArgTwo 1) 1))
        ((FullWordP FullWordP)  (!*SS1 MVC ArgOne (ByteLoc ArgTwo 3) 1))
        ((FullWordP RegP)       (STCM ArgTwo 2#0001 ArgOne))
                                                        % tag from low byte


        ((AnyP      ByteInumP)  (PROBLEM "MkItem" ArgOne ArgTwo))
        ((AnyP      ByteP)      (PROBLEM "MkItem" ArgOne ArgTwo))
        ((AnyP      HalfWordP)  (PROBLEM "MkItem" ArgOne ArgTwo))
        ((AnyP      FullWordP)  (PROBLEM "MkItem" ArgOne ArgTwo))
        ((AnyP      RegP)       (PROBLEM "MkItem" ArgOne ArgTwo))

        % misc
        (                       (!*Move ArgTwo (RegNotUsedBy ArgOne))
                                (!*MkItem ArgOne (RegNotUsedBy ArgOne))))

(DefCMacro !*Field     % dest source start size  -- do not extend sign
                        % ArgOne should not be in ArgTwo

        % tag into reg from reg or fullword
        ((RegP RegP ZeroP EightP)
                (!*Move ArgTwo ArgOne)      % load dest reg (no-op if equal)
                (SRL ArgOne 24)             % logical right shift 24 bits
                (SameLineComment ArgOne ":= tag of" ArgTwo))

        % info into reg from reg
        ((RegP RegP EightP TwentyFourP)
                (!*Move ArgTwo ArgOne)  % move entire word (no-op if equal)
                (ICM ArgOne 2#1000 (Lit (Byte 0))) % then high byte := 0
                (SameLineComment ArgOne ":= inf of" ArgTwo))

        %  big offsets (assume BigOffsetP implies FullWordP)
        ((AnyP BigOffsetP AnyP AnyP)        % reduce complexity
                (!*Move ArgTwo (RegNotUsedBy ArgOne))
                (!*Field ArgOne (RegNotUsedBy ArgOne) ArgThree ArgFour))

        ((BigOffsetP AnyP AnyP AnyP)        % reduce complexity
                (!*Field (RegNotUsedBy ArgOne) ArgTwo ArgThree ArgFour)
                (!*Move (RegNotUsedBy ArgOne) ArgOne))

        % tag into reg from fullword
        ((RegP FullWordP ZeroP EightP)      % ArgOne should not be in ArgTwo
                (SR ArgOne ArgOne)          % zero out register
                (IC ArgOne ArgTwo)          % grab first byte of word
                (SameLineComment ArgOne ":= tag of" ArgTwo))

        % info into reg from fullword
        ((RegP FullWordP EightP TwentyFourP)
                (!*Move ArgTwo ArgOne)             % dest := source
                (ICM ArgOne 2#1000 (Lit (Byte 0))) % then high byte := 0
                        %   (SR ArgOne ArgOne)     % dest := 0
                        %   (ICM ArgOne 2#0111 (ThreeByteLoc ArgTwo 1))
                                                   % low 3 bytes
                (SameLineComment ArgOne ":= inf of" ArgTwo))

        % low bits of top byte into reg from fullword
        ((RegP FullWordP LeqEightP AndSumEqualEightP)
                (IC ArgOne ArgTwo)
                (SameLineComment ArgOne ":= top byte of" ArgTwo)
                (!*WAnd ArgOne (OneByteBitMask ArgThree ArgFour))
                (SameLineComment "Mask out unwanted bits"))

        % high bits of top byte into reg from fullword
        ((RegP FullWordP ZeroP LeqEightP)
                (SR ArgOne ArgOne)
                (IC ArgOne ArgTwo)
                (SameLineComment ArgOne ":= top byte of" ArgTwo)
                (SRL ArgOne (Difference 8 ArgFour))
                (SameLineComment "Shift out unwanted bits"))

        % any other field into reg from anywhere
        ((RegP AnyP StartingBitP BitLengthP)
                (!*Move ArgTwo ArgOne)     % load dest reg (no-op if equal)
                (!*WShift ArgOne ArgThree) % shift out high bits (no-op if 0)
                (!*WShift ArgOne (Difference ArgFour BitsPerWord))
                (SameLineComment ArgOne ":= field of" ArgTwo))


        % tag into byte from reg or fullword
        ((ByteP RegP ZeroP EightP)
                (STC ArgTwo ArgOne)             % deposit tag directly
                (SameLineComment ArgOne ":= tag of" ArgTwo))
        ((ByteP FullWordP ZeroP EightP)
                (!*SS1 MVC ArgTwo ArgOne 1)     % move tag byte directly
                (SameLineComment ArgOne ":= tag of" ArgTwo))

        % all other tags (via register)
        ((AnyP AnyP ZeroP EightP)
                (!*Field (RegNotUsedBy ArgOne ArgTwo) ArgTwo 0 8)
                (!*Move (RegNotUsedBy ArgOne ArgTwo) ArgOne)
                (SameLineComment ArgOne ":= tag of" ArgTwo))

        % info into fullword from reg or fullword
        ((FullWordP RegP EightP TwentyFourP)
                (!*Move ArgTwo ArgOne)            % move entire word
                (!*SS1 MVC ArgOne (Lit (Byte 0)) 1)  % then high byte := zero
                (SameLineComment ArgOne ":= inf of" ArgTwo))
        ((FullWordP FullWordP EightP TwentyFourP)
                (!*Move ArgTwo ArgOne)            % move entire word
                (!*SS1 MVC ArgOne (Lit (Byte 0)) 1) % then high byte := zero
                (SameLineComment ArgOne ":= inf of" ArgTwo))

        % default
        ((AnyP AnyP StartingBitP BitLengthP)
               (!*Field (RegNotUsedBy ArgOne ArgTwo) ArgTwo ArgThree ArgFour)
               (!*Move (RegNotUsedBy ArgOne ArgTwo) ArgOne)
               (SameLineComment ArgOne ":= field of" ArgTwo)
               )

        (       (PROBLEM "Field" ArgOne ArgTwo ArgThree ArgFour)))


(DefCMacro !*SignedField        % dest source start size  -- extend sign
        % tag into reg from reg
        ((RegP RegP ZeroP EightP)
                (!*Move ArgTwo ArgOne)          % load destination reg
                (SRA ArgOne 24)                 % arith right shift 24 bits
                (SameLineComment ArgOne ":= signed tag of" ArgTwo))

        % info into reg from reg
        ((RegP RegP EightP TwentyFourP)
                (!*Move ArgTwo ArgOne)    % move entire word (no-op if equal)
                (SLL ArgOne 8)            % shift left 8 bits
                (SameLineComment " Shift top bit up to sign bit")
                (SRA ArgOne 8)            % then back, extending sign
                (SameLineComment " extend sign to get signed inf"))

        %  big offsets (assume BigOffsetP implies FullWordP)
        ((AnyP BigOffsetP AnyP AnyP)
                (!*Move ArgTwo (RegNotUsedBy ArgOne))   % reduce complexity
                (!*SignedField ArgOne (RegNotUsedBy ArgOne) ArgThree ArgFour) )

        ((BigOffsetP AnyP AnyP AnyP)
                (!*SignedField (RegNotUsedBy ArgOne) ArgTwo ArgThree ArgFour)
                (!*Move (RegNotUsedBy ArgOne) ArgOne))


        % tag into reg from fullword
        ((RegP FullWordP ZeroP EightP)
                (!*Move ArgTwo ArgOne)          % load destination reg
                (SRA ArgOne 24)                 % arith right shift 24 bits
                (SameLineComment ArgOne ":= signed tag of" ArgTwo))

        % inf into reg from fullword
        ((RegP FullWordP EightP TwentyFourP)
                (ICM ArgOne 2#1110 (ThreeByteLoc ArgTwo 1))
                                                % move to high 3 bytes
                (SameLineComment ArgOne ":= inf of" ArgTwo "shifted up")
                (SRA ArgOne 8)                  % then right, extending sign
                (SameLineComment "  then extend sign down"))

        % other into reg from any
        ((RegP AnyP StartingBitP BitLengthP)
                (!*Move ArgTwo ArgOne)     % move entire wd (no-op if equal)
                (!*WShift ArgOne ArgThree) % shift away high bits (noop if 0)
                (!*AShift ArgOne (Difference ArgFour BitsPerWord))
                                           % shift right, extending sign
                (SameLineComment ArgOne ":= signed field of" ArgTwo))

        % tag into byte (slightly special case)
        ((ByteP RegP ZeroP EightP)
                (STC ArgTwo ArgOne))            % deposit tag directly
        ((ByteP FullWordP ZeroP EightP)
                (!*SS1 MVC ArgTwo ArgOne 1))    % move tag byte directly
        ((ByteP HalfWordP ZeroP EightP)
                (!*SS1 MVC ArgTwo ArgOne 1))    % move tag byte directly
        ((ByteP ByteP ZeroP EightP)
                (!*SS1 MVC ArgTwo ArgOne 1))    % move tag byte directly


        % any other signed field from anywhere
        ((AnyP AnyP StartingBitP BitLengthP)
              (!*SignedField (RegNotUsedBy ArgOne) ArgTwo ArgThree ArgFour)
                                                    % get into register
              (!*Move (RegNotUsedBy ArgOne) ArgOne) % then move to dest
              (SameLineComment ArgOne ":= signed field of" ArgTwo))

        % perhaps illegal bits were specified
        (       (PROBLEM "SignedField" ArgOne ArgTwo ArgThree ArgFour)))

% Note that the arguments to !*PutField are source, destination, start, size.
% This is not what the CMACRO document indicates. MkItem above has been
% modified to reflect this ordering.  This is entirely backwards to the
% majority of the CMACRO's....

(DefCMacro !*PutField                   % source, dest, start, size
        ((AnyP AnyP ZeroP EightP)
           (!*MkItem ArgTwo ArgOne))

        % various things into Register

        ((FullInumP RegP EightP TwentyFourP)
           (ICM ArgOne 2#0111   % insert into 3 low order bytes of register
                (ThreeByteLoc   % low three bytes
                     (lit (fullword (hex (LAnd ArgOne 16#FFFFFF))))
                     1)))

        ((FullInumP RegP StartingBitP BitLengthP)
           (!*WAnd ArgTwo (LNot (BitMask ArgThree ArgFour)))
           (SameLineComment "Zero out field")
           (!*WOr ArgTwo
                  (lsh (LAnd ArgOne (Sub1 (Expt 2 ArgFour)))
                       (difference BitsPerWord (plus ArgThree ArgFour))))
           (SameLineComment "Put new value in field"))

        % various things into FullWord

        ((RegP FullwordP EightP TwentyFourP)
                (STCM ArgOne 2#0111 (ThreeByteLoc ArgTwo 1)))

        ((FullInumP FullwordP EightP TwentyFourP)
                (!*Move ArgOne (RegNotUsedBy ArgTwo))
                (STCM (RegNotUsedBy ArgTwo) 2#0111 (ThreeByteLoc ArgTwo 1)))

        ((FullInumP FullwordP LeqEightP AndSumLeqEightP)
                (IC (RegNotUsedBy ArgOne ArgTwo) ArgTwo)
                (SameLineComment "Extract byte")
                (!*WAnd (RegNotUsedBy ArgOne ArgTwo)
                        (LXor 16#FF (OneByteBitMask ArgThree ArgFour)))
                (SameLineComment "Zero out field")
                (!*WOr (RegNotUsedBy ArgOne ArgTwo)
                       (lsh (LAnd ArgOne (Sub1 (Expt 2 ArgFour)))
                            (difference BitsPerByte
                                        (plus ArgThree ArgFour))))
                (SameLineComment "Put new value in field")
                (STC (RegNotUsedBy ArgOne ArgTwo) ArgTwo)
                (SameLineComment "Reinsert byte"))

        ((AnyP FullwordP LeqEightP AndSumLeqEightP)
                (!*Move ArgOne (RegNotUsedBy ArgOne ArgTwo))
                (SameLineComment "Get byte to insert")
                (!*WShift (RegNotUsedBy ArgOne ArgTwo)
                          (difference BitsPerByte
                                      (plus ArgThree ArgFour)))
                (SameLineComment "Shift data into position")
                (!*WAnd (RegNotUsedBy ArgOne ArgTwo)
                        (OneByteBitMask ArgThree ArgFour))
                (SameLineComment "Zero out non-field bits")
                (IC (SecondRegNotUsedBy ArgOne ArgTwo) ArgTwo)
                (SameLineComment "Get byte to be altered")
                (!*WAnd (SecondRegNotUsedBy ArgOne ArgTwo)
                        (LXOr 16#FF (OneByteBitMask ArgThree ArgFour)))
                (SameLineComment "Zero out field bits")
                (OR (RegNotUsedBy ArgOne ArgTwo)
                    (SecondRegNotUsedBy ArgOne ArgTwo))
                (SameLineComment "Put new value in field")
                (STC (RegNotUsedBy ArgOne ArgTwo) ArgTwo)
                (SameLineComment "Reinsert byte"))

        ((AnyP AnyP StartingBitP BitLengthP)
          (Comment "Move " ArgOne " into field in " ArgTwo)
          (!*Move ArgOne (RegNotUsedBy ArgTwo))
          (!*WShift (RegNotUsedBy ArgTwo)
                    (difference BitsPerWord (plus ArgThree ArgFour)))
          (!*WAnd (RegNotUsedBy ArgTwo) (BitMask ArgThree ArgFour))
          (!*Move ArgTwo (SecondRegNotUsedBy ArgTwo))
          (!*WAnd (SecondRegNotUsedBy ArgTwo)
                  (LNot (BitMask ArgThree ArgFour)))
          (!*WOr (SecondRegNotUsedBy ArgTwo) (RegNotUsedBy ArgTwo))
          (!*Move (SecondRegNotUsedBy ArgTwo) ArgTwo))

        % bit positions must be 0 to 36
        (       (PROBLEM "PutField" ArgOne ArgTwo ArgThree ArgFour)))

(DefCMacro !*Loc       % arg1 := location of arg2
        ((AnyP      RegP)  (PROBLEM "Loc" ArgOne ArgTwo))
        ((RegP      AnyP)  (LA ArgOne ArgTwo))
        ((FullWordP AnyP)  (LA (RegNotUsedBy ArgOne) ArgTwo)
                           (ST (RegNotUsedBy ArgOne) ArgOne))
        (                  (PROBLEM "Loc" ArgOne ArgTwo)))
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                       MISC. CMACROS                                   %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(DefCMacro !*JumpOn)
(DefCMacro !*ForeignLink)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Portable cmacro definitions *LAMBIND, *PROGBIND and *FREERSTR         %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(DefCMacro !*LamBind)

(DefCMacro !*ProgBind)

(DefCMacro !*FreeRstr)

(setq !*UnsafeBinder T)                 % has to save registers across calls

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
(DefList '((Byte        ((SameLineComment "R1 pts at base of array")
                         (AR (Reg 2) (Reg 1))
                         (SameLineComment "R2 := location of byte")
                         (SR (Reg 1) (Reg 1))
                         (IC (Reg 1) (Indexed (Reg 2) 0))
                         (SameLineComment "R1 := byte from array")))
           (PutByte     ((SameLineComment "R1 pts at base of array")
                         (STC (Reg 3) (DoubleIndexed (Reg 1) (Reg 2) 0))
                         (SameLineComment "Deposit byte in array")))
           (HalfWord    ((SameLineComment "R1 pts at base of array")
                         (SLL (Reg 2) 1)
                         (SameLineComment "Byte offset := 2 * HW offset")
                         (LH  (Reg 1) (DoubleIndexed (Reg 1) (Reg 2) 0))
                         (SameLineComment "R1 := halfword from array")))
           (PutHalfWord ((SameLineComment "R1 pts at base of array")
                         (SLL (Reg 2) 1)
                         (SameLineComment "Byte offset := 2 * HW offset")
                         (STH (Reg 3) (DoubleIndexed (Reg 1) (Reg 2) 0))
                         (SameLineComment "Deposit halfword in array")))
           )
         'OpenCode)

(!&TwoReg '(Byte PutByte HalfWord PutHalfWord))


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                          %
%  All fn calls are via CallReg, except exit calls which use BaseR and     %
%  enter ten byes into the called function.                                %
%  If started normally, the called routine loads BaseR from CallReg to     %
%  get addressibility.                                                     %
%  If entered down ten bytes, the routine assumes it already has           %
%  addressibility from BaseR (i.e. BaseR points at the START of the fn).   %
%                                                                          %
%  See the section on function calls.                                      %
%                                                                          %
%  All calls set RetR so CCI and NULLDEFN can deduce the ID number.        %
%                                                                          %
%  In the following, the arguments are in C1-Cn, and the fn is determined  %
%  by Cn+1, either as an ID number or as a direct address.                 %
%                                                                          %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(DefList
 '((IDApply0 ((SLL (Reg 1) 2)
              (LR (Reg LinkReg) (Reg 1))   % ID offset
              (L (Reg CallReg) (DoubleIndexed (Reg 1) (Reg SymFncR) 0))
              (BALR (Reg RetR) (Reg CallReg))))
   (IDApply1 ((SLL (Reg 2) 2)
              (LR (Reg LinkReg) (Reg 2))   % ID offset
              (L (Reg CallReg) (DoubleIndexed (Reg 2) (Reg SymFncR) 0))
              (BALR (Reg RetR) (Reg CallReg))))
   (IDApply2 ((SLL (Reg 3) 2)
              (LR (Reg LinkReg) (Reg 3))   % ID offset
              (L (Reg CallReg) (DoubleIndexed (Reg 3) (Reg SymFncR) 0))
              (BALR (Reg RetR) (Reg CallReg))))
   (IDApply3 ((SLL (Reg 4) 2)
              (LR (Reg LinkReg) (Reg 4))   % ID offset
              (L (Reg CallReg) (DoubleIndexed (Reg 4) (Reg SymFncR) 0))
              (BALR (Reg RetR) (Reg CallReg))))
   (IDApply4 ((SLL (Reg 5) 2)
              (LR (Reg LinkReg) (Reg 5))   % ID offset
              (L (Reg CallReg) (DoubleIndexed (Reg 5) (Reg SymFncR) 0))
              (BALR (Reg RetR) (Reg CallReg))))
           )
         'OpenCode)

(DefList '(
   (IDApply0 ((SLL (Reg 1) 2)
              (LR (Reg LinkReg) (Reg 1))   % ID offset
              (L (Reg BaseR) (DoubleIndexed (Reg 1) (Reg SymFncR) 0))
              (BAL (Reg RetR) (Indexed (Reg BaseR) 10))))
   (IDApply1 ((SLL (Reg 2) 2)
              (LR (Reg LinkReg) (Reg 2))   % ID offset
              (L (Reg BaseR) (DoubleIndexed (Reg 2) (Reg SymFncR) 0))
              (BAL (Reg RetR) (Indexed (Reg BaseR) 10))))
   (IDApply2 ((SLL (Reg 3) 2)
              (LR (Reg LinkReg) (Reg 3))   % ID offset
              (L (Reg BaseR) (DoubleIndexed (Reg 3) (Reg SymFncR) 0))
              (BAL (Reg RetR) (Indexed (Reg BaseR) 10))))
   (IDApply3 ((SLL (Reg 4) 2)
              (LR (Reg LinkReg) (Reg 4))   % ID offset
              (L (Reg BaseR) (DoubleIndexed (Reg 4) (Reg SymFncR) 0))
              (BAL (Reg RetR) (Indexed (Reg BaseR) 10))))
   (IDApply4 ((SLL (Reg 5) 2)
              (LR (Reg LinkReg) (Reg 5))   % ID offset
              (L (Reg BaseR) (DoubleIndexed (Reg 5) (Reg SymFncR) 0))
              (BAL (Reg RetR) (Indexed (Reg BaseR) 10))))
          ) 'ExitOpenCode)

(DefList '(
        (AddressApply0 ((LR (Reg CallReg) (Reg 1))
                        (BALR (Reg RetR) (Reg CallReg))))
        (AddressApply1 ((LR (Reg CallReg) (Reg 2))
                        (BALR (Reg RetR) (Reg CallReg))))
        (AddressApply2 ((LR (Reg CallReg) (Reg 3))
                        (BALR (Reg RetR) (Reg CallReg))))
        (AddressApply3 ((LR (Reg CallReg) (Reg 4))
                        (BALR (Reg RetR) (Reg CallReg))))
        (AddressApply4 ((LR (Reg CallReg) (Reg 5))
                        (BALR (Reg RetR) (Reg CallReg))))
          ) 'OpenCode)

(DefList '(
        (AddressApply0 ((LR (Reg BaseR) (Reg 1))
                        (BAL (Reg RetR) (Indexed (Reg BaseR) 10))))
        (AddressApply1 ((LR (Reg BaseR) (Reg 2))
                        (BAL (Reg RetR) (Indexed (Reg BaseR) 10))))
        (AddressApply2 ((LR (Reg BaseR) (Reg 3))
                        (BAL (Reg RetR) (Indexed (Reg BaseR) 10))))
        (AddressApply3 ((LR (Reg BaseR) (Reg 4))
                        (BAL (Reg RetR) (Indexed (Reg BaseR) 10))))
        (AddressApply4 ((LR (Reg BaseR) (Reg 5))
                        (BAL (Reg RetR) (Indexed (Reg BaseR) 10))))
          ) 'ExitOpenCode)

% This is an optimization to be added later.
% (!&Tworeg '(WQuotient WRemainder Byte PutByte))

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%               370 STANDARD INSTRUCTION SET                                %
%                                                                           %
%  Arithmetic:                                                              %
%                                                                           %
%  +    Add:            AR, A, AH, ALR, AL                                  %
%  +    Subtract:       SR, S, SH, SLR, SL                                  %
%  +    Mulitply:       MR, M, MH                                           %
%  +    Divide:         DR, D                                               %
%                                                                           %
%  Logical:                                                                 %
%                                                                           %
%  +    And             NR, N, NC, NI                                       %
%  +    Or              OR, O, OC, OI                                       %
%  +    Xor             XR, X, XC, XI                                       %
%  +    Not             LCR                                                 %
%                                                                           %
%  Shifts:                                                                  %
%                                                                           %
%  +    Arithmetic      SLA, SLDA, SRA, SRDA                                %
%  +    Logical         SLL, SLDL, SRL, SRDL                                %
%                                                                           %
%  Branch/Link:                                                             %
%                                                                           %
%  +    Call            BALR, BAL                                           %
%  +    Branch          BCR, BC, BCTR, BCT, BXH, BXLE                       %
%  +    Interlock       TS,  BC, BCTR, BCT, BXH, BXLE                       %
%                                                                           %
%                                                                           %
%  Comparison:                                                              %
%                                                                           %
%  +    Arithmetic      CR, C, CH                                           %
%  +    Logical         CLR, CL, CLC, CLI, CLM, CLCL, TM                    %
%                                                                           %
%  Movement:                                                                %
%                                                                           %
%  +    Load            LR, L, LA, LH, LM, LNR, LPR, IC, ICM                %
%  +    Load & test     LTR                                                 %
%  +    Store           ST, STH, STM, STC, STCM                             %
%  +    Mem-to-Mem      MVC, MVI, MVCL, MVN, MVO, MVZ                       %
%                                                                           %
%                                                                           %
%  IO / Conversions                                                         %
%                                                                           %
%       IO              CLRIO, SIO, SIOF, HIO, HDV, TCH, TIO                %
%       Conversions     CVB, CVD, PACK, UNPK, TR, TRT                       %
%                                                                           %
%  System info                                                              %
%                                                                           %
%       Supervisor      MC, SVC, Diagnose                                   %
%       Store           STCK, STCTL, STIDP, STIDC, ISK                      %
%       Set             LPSW, SCK, SPM, SSK, SSM, LCTL                      %
%                                                                           %
%  Other                                                                    %
%                                                                           %
%       Execute         EX                                                  %
%                                                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                    DECIMAL FEATURE INSTRUCTIONS                           %
%                                                                           %
%    Arithmetic:        AP, MP, DP, SP, ZAP                                 %
%    Shift:             SRP                                                 %
%  + Compare:           CP                                                  %
%    Conversion:        ED, EDMK                                            %
%                                                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                    FLOATING POINT FEATURE INSTRUCTIONS                    %
%                                                                           %
%  Arithmetic:                                                              %
%                                                                           %
%    Add                ADR, AD, AER, AE, AWR, AW, AUR, AU                  %
%    Subtract           SDR, SD, SER, SE, SWR, SW, SUR, SU                  %
%    Multiply           MDR, MD, MER, ME                                    %
%    Divide             DDR, DD, DER, DE, HDR, HER                          %
%  + Compare            CDR, CD, CER, CE                                    %
%    Load               LDR, LD, LER, LE, LTDR, LTER,                       %
%                       LCDR, LCER, LNDR, LNER, LPDR, LPER                  %
%    Store              STD, STE                                            %
%                                                                           %
%                                                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%       EXTENDED-PRECISION FLOATING-POINT FEATURE INSTRUCTIONS              %
%                                                                           %
%    Arith              AXR, SXR, MXR, MXDR, MXD                            %
%    Load               LRDR, LRER                                          %
%                                                                           %
%                                                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%               OTHER SPECIAL INSTRUCTIONS                                  %
%                                                                           %
%    CPU TIMER                  SCKC, SPT, STCKC, STPT                      %
%    DIRECT-CONTROL             RDD, WRD                                    %
%    TRANSLATION-FEATURE        LRA, PTLB, RRB, STNSM, STOSM                %
%    MULTIPROCESSING            SPX, SIGP, STAP, STPX                       %
%    CONDITIONAL-SWAPPING       CS, CDS                                     %
%    PSW KEY-HANDLING           IPK, SPKA                                   %
%                                                                           %
%                                                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%       that's all folks!!                                                  %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
